-- -*-haskell-*-

-- Gpc.chs

-- HGPC: A Haskell port for GPC.
-- This file is part of HGPC.
-- Copyright (C) 2007 Marco Túlio Gontijo e Silva
-- Copyright (C) 2007 Rafael Cunha de Almeida

-- HGPC is free software; you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2 of the License, or
-- (at your option) any later version.

-- HGPC is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.

-- You should have received a copy of the GNU General Public License
-- along with HGPC; if not, write to the Free Software
-- Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

#include <gpcl/gpc.h>

-- | HGPC: A Haskell port for GPC.
--
-- HGPC is a Haskell port for the General Polygon Clipper Library (GPC),
-- a flexible and highly robust polygon set operations library. GPC is free for
-- non-comercial use and can be obtained at
-- <http://www.cs.man.ac.uk/~toby/alan/software/>.

module Algebra.Geometric.Gpc (
        -- * Geometric data types
        Polygon (PolygonC, polygonSet),
        Contour (ContourC, contourList),
        Vertex (VertexC, vertexX, vertexY),
        -- * Polygon functions
        -- ** Unsafe functions
        difference, intersection, xor, union,
        -- ** Safe functions
        safeDifference, safeIntersection, safeXor, safeUnion)
    where

import Prelude hiding (null)

import Data.List hiding (union, null)
import Data.Set hiding (difference, intersection, union, map)
import Control.Monad
import Foreign hiding (xor)
import Foreign.C

{#context prefix = "gpc"#}

{#pointer *polygon as CPolygon -> Polygon#}
{#pointer *vertex_list as CContour -> Contour#}
{#pointer *vertex as CVertex -> Vertex#}

-- | Returns a 'Polygon' with the area in the first 'Polygon' and not in the
-- second.
difference :: Polygon -> Polygon -> Polygon
difference subject clip = unsafePerformIO $ safeDifference subject clip

-- | Returns a 'Polygon' with the area in both the first and the second
-- 'Polygon'.
intersection :: Polygon -> Polygon -> Polygon
intersection subject clip = unsafePerformIO $ safeIntersection subject clip

-- | Returns a 'Polygon' with the area in the first or the second 'Polygon',
-- but not in both.
xor :: Polygon -> Polygon -> Polygon
xor subject clip = unsafePerformIO $ safeXor subject clip

-- | Returns a 'Polygon' with the area in the first or the second 'Polygon'.
union :: Polygon -> Polygon -> Polygon
union subject clip = unsafePerformIO $ safeUnion subject clip

-- | A 'Polygon', which support holes and disjoint areas. Each 'Bool' in the
-- tuple tells if the 'Contour' is the border of a hole ('True') or of a
-- 'Polygon' ('False').
newtype Polygon = PolygonC {polygonSet :: Set (Bool, Contour)} deriving Show

instance Storable Polygon where
    sizeOf _ = {#sizeof polygon#}

    alignment _ = intAlignment

    peek cPolygon =
        (PolygonC . fromList) `liftM`
        do
        numContours <-
            fromIntegral `liftM` {#get polygon.num_contours#} cPolygon

        liftM2 zip (

                map (toEnum . fromIntegral) `liftM`
                ({#get polygon.hole#} cPolygon >>= peekArray numContours)) $

            {#get polygon.contour#} cPolygon >>= peekArray numContours

    poke cPolygon (PolygonC polygon) =
        {#set polygon.num_contours#} cPolygon (fromIntegral $ size polygon)

        >> newArray (map (fromIntegral . fromEnum) holes) >>=
        {#set polygon.hole#} cPolygon >>

        newArray contours >>= {#set polygon.contour#} cPolygon

        where (holes, contours) = unzip $ toList polygon

-- | A list of 2D Points. A 'Contour' is the border of a 'Polygon' or the
-- internal border of a hole in a 'Polygon'.
newtype Contour = ContourC {contourList :: [Vertex]} deriving (Ord, Show)

instance Eq Contour where
    (ContourC list1) == (ContourC list2) =
        list1 `elem` rotation list2 ++ rotation (reverse list2)

instance Storable Contour where
    sizeOf _ = {#sizeof vertex_list#}

    alignment _ = intAlignment

    peek cContour =
        ContourC `liftM`
        do
        numVertices <-
            fromIntegral `liftM` {#get vertex_list.num_vertices#} cContour

        {#get vertex_list.vertex#} cContour >>= peekArray numVertices

    poke cContour (ContourC contour) =
        {#set vertex_list.num_vertices#} cContour
        (fromIntegral $ length contour) >>

        newArray contour >>= {#set vertex_list.vertex#} cContour

-- | A 2D Point.
data Vertex = VertexC {
    vertexX :: Double,
    vertexY :: Double
    }
    deriving (Eq, Ord, Show)

instance Storable Vertex where
    sizeOf _ = {#sizeof vertex#}

    alignment _ = intAlignment

    peek cVertex =
        liftM2 VertexC (get {#get vertex.x#}) $ get {#get vertex.y#}
        where get field = realToFrac `liftM` field cVertex

    poke cVertex (VertexC x y) =
        {#set vertex.x#} cVertex (realToFrac x) >>
        {#set vertex.y#} cVertex (realToFrac y)

-- | An 'IO' version of 'difference', which does not use 'unsafePerformIO'.
safeDifference :: Polygon -> Polygon -> IO Polygon
safeDifference subject clip = polygonClip 0 subject clip

-- | An 'IO' version of 'intersection', which does not use 'unsafePerformIO'.
safeIntersection :: Polygon -> Polygon -> IO Polygon
safeIntersection subject clip = polygonClip 1 subject clip

-- | An 'IO' version of 'xor', which does not use 'unsafePerformIO'.
safeXor :: Polygon -> Polygon -> IO Polygon
safeXor subject clip = polygonClip 2 subject clip

-- | An 'IO' version of 'union', which does not use 'unsafePerformIO'.
safeUnion :: Polygon -> Polygon -> IO Polygon
safeUnion subject clip = polygonClip 3 subject clip

rotation :: Eq a => [a] -> [[a]]
rotation list = nub $ zipWith (++) (tails list) $ inits list

intAlignment :: Int
intAlignment = alignment (undefined :: Int)

polygonClip :: CInt -> Polygon -> Polygon -> IO Polygon
polygonClip op subject clip =
    alloca $
    \cSubject -> alloca $
    \cClip -> alloca $
    \cResult ->
    do
    poke cSubject subject
    poke cClip clip

    {#call polygon_clip#} op cSubject cClip cResult
    result <- peek cResult

    {#call free_polygon#} cSubject
    {#call free_polygon#} cClip
    {#call free_polygon#} cResult

    return result
